/**
* @author Dirk Nordmann
* @date   14.01.2002
*/
package xml2rtf.rtf;

import util.Globals;
import xml2rtf.xml.*;
import xml2rtf.templates.*;
import java.util.*;

public class RtfTable
{
    private class Cell
    {
        private int width;
        private boolean percentWidth;
        private int longestNonBreaking;
        private int longestLine;
        private int currentLineLength;
        private int colSpan;
        private int rowSpan;
        private int _align;
        private boolean rowSpanSpacer;
        private String _bgcolor;
        private boolean isHeader;
        private StringBuffer contents;
        private int textSize;

        /**
        * Constructor
        */
        public Cell()
        {
            percentWidth = false;
            colSpan = 1;
            rowSpan = 1;
            _align = RtfGeneratorContext.ALIGN_LEFT;
            rowSpanSpacer = false;
            _bgcolor = "";
            isHeader = false;
            contents = new StringBuffer();
        }
    }

    private class Row
    {

        private int _align;
        private int cellNumber;
        private String _bgcolor;
        private Vector _cells;

        /**
        * Constructor
        */
        public Row()
        {
            _align = RtfGeneratorContext.ALIGN_LEFT;
            _bgcolor = "";
            _cells = new Vector();
        }

        void addCell(Cell cell)
        {
        		_cells.addElement(cell);
        }

        Cell getCell(int i)
        {
						return (Cell)_cells.elementAt(i);
    		}

    		int cellCount()
    		{
    				return _cells.size();
    		}

    }



/*	######################################################################
	###	CONSTRUCTOR
	###
	###	IN:  	Tag					Tag-Object
	###			initial_offset		Initial offset
	###			default_align		Default alignment
	###			RtfWriter			Instance of RTF-Writer
	###	OUT: 	RtfTable			New instance of Tables
	######################################################################
 */
    public RtfTable(Tag tag, int initial_offset, int default_align, RtfWriter rtfwriter)
    {
        rows = new Vector();
        widthInTwips = ConversionTemplate.maxTwipsWidth;
        explicitTableSize = false;
        dumpToCurrentRow = false;
        output = rtfwriter;

        //* Table border *
        String border = tag.getTagAttr("border");
        if (border == null)
            hasBorder = false;
        else
            hasBorder = !border.equals("0");

        //* Initial offset *
        initialOffset = initial_offset;

        //* Width *
        int width = tag.getNumericTagAttr("width");
        if(width == Tag.NO_VALUE)
        {
            String sWidth = tag.getTagAttr("width");
            if(sWidth != null && sWidth.endsWith("%"))
            {
                sWidth = sWidth.substring(0, sWidth.length() - 1);
                try
                {
                    widthInTwips = (ConversionTemplate.maxTwipsWidth * Integer.parseInt(sWidth)) / 100;
                    explicitTableSize = true;
                }
                catch(NumberFormatException _ex) { }
            }
        } else
        {
            explicitTableSize = true;
            widthInTwips = (int)(width * Globals.PIX2TWIPS);
            if(widthInTwips > ConversionTemplate.maxTwipsWidth)
                widthInTwips = ConversionTemplate.maxTwipsWidth;
        }

        //* Alignment *
        String align = tag.getTagAttr("align");
        if (align != null)
        {
            align = align.toUpperCase();
            if(align.equals("RIGHT"))
                _align = RtfGeneratorContext.ALIGN_RIGHT;
            else
            if(align.equals("CENTER"))
                _align = RtfGeneratorContext.ALIGN_CENTER;
            else
                _align = RtfGeneratorContext.ALIGN_LEFT;
        } else
        {
            _align = default_align;
        }

        //* BackgroundColor *
        _bgcolor = getColor(tag);

        widthInTwips -= initialOffset;
    }



    private int getAlign(Tag tag)
    {
        int rtn = currentRow == null ? RtfGeneratorContext.ALIGN_LEFT : currentRow._align;
        String align = tag.getTagAttr("align");
        if(align != null)
        {
            align = align.toUpperCase();
            if(align.equals("RIGHT"))
                rtn = RtfGeneratorContext.ALIGN_RIGHT;
            else
            if(align.equals("CENTER"))
                rtn = RtfGeneratorContext.ALIGN_CENTER;
        }
        return rtn;
    }



    private String getColor(Tag tag)
    {
    		return getColor(tag,"");
    }



    private String getColor(Tag tag, String s1)
    {
        String rtn = s1;
        int bgcolor = tag.getColorAttr("bgcolor");
        if(bgcolor != Tag.NO_VALUE)
            rtn += output.getColor("clcbpat",bgcolor) + " ";
        return rtn;
    }



    public void storeCurrentCell()
    {
        if(currentCell != null && currentRow != null)
        {
            insertLineBreak();
            currentRow.addCell(currentCell);
            currentCell = null;
        }
    }



    public void storeCurrentRow(boolean flag)
    {
        if(currentRow != null)
        {
            storeCurrentCell();
            for(int i = currentRow.cellNumber; i < maxCellPerRow; i++)
                checkRowSpanning(i);

            if(currentRow.cellCount() > 0)
            {
                rows.addElement(currentRow);
                if(currentRow.cellCount() > maxCellPerRow)
                    maxCellPerRow = currentRow.cellCount();
            }
            if(flag)
                currentRow = null;
        }
    }



    private void checkRowSpanning(int i)
    {
        int j = 0;
        for(int k = rows.size() - 1; k >= 0; k--)
        {
            Row row = (Row)rows.elementAt(k);
            int l = 0;
            for(int i1 = 0; i1 < row.cellCount(); i1++)
            {
                Cell cell = row.getCell(i1);
                if(i == l && cell.rowSpan - j > 1)
                {
                    currentCell = new Cell();
                    currentCell.rowSpanSpacer = true;
                    currentCell._bgcolor = cell._bgcolor;
                    storeCurrentCell();
                }
                l += cell.colSpan;
            }

            j++;
        }

    }



    public void addCell(Tag tag, boolean flag)
    {
        storeCurrentCell();
        checkRowSpanning(currentRow.cellNumber);
        currentCell = new Cell();
        currentCell.isHeader = flag;
        currentCell._align = getAlign(tag);
        if((currentCell.colSpan = tag.getNumericTagAttr("colspan")) == Tag.NO_VALUE)
            currentCell.colSpan = 1;
        currentRow.cellNumber += currentCell.colSpan;
        if((currentCell.rowSpan = tag.getNumericTagAttr("rowspan")) == Tag.NO_VALUE)
            currentCell.rowSpan = 1;
        currentCell._bgcolor = getColor(tag, currentRow._bgcolor);
        if((currentCell.width = tag.getNumericTagAttr("width")) == Tag.NO_VALUE)
        {
            String s1 = tag.getTagAttr("width");
            if(s1 != null && s1.endsWith("%"))
            {
                s1 = s1.substring(0, s1.length() - 1);
                try
                {
                    currentCell.width = (widthInTwips * Integer.parseInt(s1)) / 100;
                }
                catch(NumberFormatException _ex) { }
            }
        } else
            currentCell.width *= Globals.PIX2TWIPS;
    }



    public void addRow(Tag tag)
    {
        storeCurrentRow(true);
        if(dumpToCurrentRow)
        {
            dumpToCurrentRow = false;
            dumpRTF();
            rows.removeAllElements();
            currentRow = null;
            currentCell = null;
        }
        currentRow = new Row();
        currentRow._align = getAlign(tag);
        currentRow._bgcolor = getColor(tag, _bgcolor);
    }



    public void addTextToCell(String s)
    {
        if(currentCell != null)
            currentCell.contents.append(s);
    }



    public void addTextContentsToCell(String text, int jfontsize, boolean bold, boolean italic, boolean fixed)
    {
        if(currentCell != null)
        {
            if(currentCell.isHeader)
                bold = true;
            int j = textSizeInTwips(text, jfontsize, bold, italic, fixed);
            currentCell.textSize += j;
            currentCell.currentLineLength += j;
            for(StringTokenizer stringtokenizer = new StringTokenizer(text, " \t\n"); stringtokenizer.hasMoreTokens();)
            {
                int k = textSizeInTwips(stringtokenizer.nextToken(), jfontsize, bold, italic, fixed);
                if(k > currentCell.longestNonBreaking)
                    currentCell.longestNonBreaking = k;
            }
        }
    }



    public void insertLineBreak()
    {
        if(currentCell != null)
        {
            if(currentCell.currentLineLength > currentCell.longestLine)
                currentCell.longestLine = currentCell.currentLineLength;
            currentCell.currentLineLength = 0;
        }
    }



    public void addTableToCell(RtfTable rtftable)
    {
        rtftable.storeCurrentRow(true);
        currentCell = null;
        Enumeration enumeration = rtftable.rows.elements();
        if(enumeration.hasMoreElements())
        {
            Row row = (Row)enumeration.nextElement();
        }
        while(enumeration.hasMoreElements())
        {
            storeCurrentRow(true);
            currentRow = (Row)enumeration.nextElement();
        }
        dumpToCurrentRow = true;
    }



    public void calculateTable()
    {
        longestNonBreaking = new int[maxCellPerRow];
        maxLineSize = new int[maxCellPerRow];
        cellSizes = new int[maxCellPerRow];
        isEmpty = new boolean[maxCellPerRow];
        explicitSize = new boolean[maxCellPerRow];
        int _tmp = widthInTwips / maxCellPerRow;
        int i1 = -1;
        for(int j1 = 0; j1 < maxCellPerRow; j1++)
        {
            longestNonBreaking[j1] = 0;
            maxLineSize[j1] = 0;
            cellSizes[j1] = 0;
            isEmpty[j1] = true;
            explicitSize[j1] = false;
        }

        for(Enumeration enumeration = rows.elements(); enumeration.hasMoreElements();)
        {
            Row row = (Row)enumeration.nextElement();
            i1++;
            int i = 0;
            for (int j = 0; j < row.cellCount(); j++)
            {
                Cell cell = row.getCell(j);
                if(cell.rowSpan > 1 && cell.width <= 0 && cell.colSpan <= 1)
                    cell.width = cell.longestLine;
                int i2;
                if(cell.width > 0 && cell.colSpan <= 1)
                {
                    i2 = cell.width;
                    if(i2 < cell.longestNonBreaking)
                        i2 = cell.longestNonBreaking;
                } else
                {
                    i2 = cell.longestNonBreaking;
                }
                i2 += MARGIN;
                try
                {
                    if(cell.longestLine > maxLineSize[i])
                        maxLineSize[i] = cell.longestLine + MARGIN;
                    if(cell.longestNonBreaking > longestNonBreaking[i])
                        longestNonBreaking[i] = cell.longestNonBreaking;
                    if(!explicitSize[i])
                        explicitSize[i] = cell.width > 0;
                    totalSize += i2;
                    if(isEmpty[i])
                        isEmpty[i] = i2 <= MARGIN;
                    if(i2 > cellSizes[i])
                    {
                        cellSizes[i] = i2;
                    }
                }
                catch(IndexOutOfBoundsException _ex)
                {
                    System.out.println("[RtfTable.calculateTable]: WARNUNG -- (supposed) nested tables are not processed.");
                }
                i += cell.colSpan;
            }

        }

        int k1 = widthInTwips;
        int l1 = maxCellPerRow;
        for(int j = 0; j < cellSizes.length; j++)
        {
            if(explicitSize[j])
            {
                k1 -= cellSizes[j];
                l1--;
            } else
            if(isEmpty[j])
                l1--;
        }

        if(k1 < 0)
        {
            cellSizes[cellSizes.length - 1] += k1;
            k1 = 0;
        }
        if(l1 > 0)
        {
            int j2 = maxCellPerRow;
            k1 /= l1;
            for(int k = 0; k < cellSizes.length; k++)
            {
                if(!isEmpty[k])
                {
                    if((cellSizes[k] <= MARGIN || !explicitSize[k]) && k1 >= longestNonBreaking[k] + MARGIN)
                        if(k1 > maxLineSize[k] && !explicitTableSize)
                        {
                            cellSizes[k] = maxLineSize[k];
                            if(j2 > 1)
                                k1 += (k1 - maxLineSize[k]) / (j2 - 1);
                        } else
                        {
                            cellSizes[k] = k1;
                        }
                } else
                {
                    cellSizes[k] = 0;
                }
                j2--;
            }

        }
        if(!explicitTableSize)
        {
            widthInTwips = 0;
            for(int l = 0; l < cellSizes.length; l++)
                widthInTwips += cellSizes[l];

            return;
        }
        int k2 = 0;
        int l2 = 0;
        int i3 = 0;
        for(int j3 = 0; j3 < cellSizes.length; j3++)
        {
            if(isEmpty[j3])
            {
                l2++;
                cellSizes[j3] = 0;
            }
            k2 += cellSizes[j3];
        }

        i3 = cellSizes.length - l2;
        if(i3 > 0 && k2 != widthInTwips)
        {
            int k3 = (widthInTwips - k2) / i3;
            for(int l3 = 0; l3 < cellSizes.length; l3++)
                if(!isEmpty[l3])
                    cellSizes[l3] += k3;

        }
    }

    public void dumpRTF()
    {
        storeCurrentRow(true);
        if(maxCellPerRow <= 0)
            return;
        calculateTable();
        if(totalSize == 0)
            return;
        boolean flag = widthInTwips == ConversionTemplate.maxTwipsWidth;
        int i = 0;
        if(!flag)
            i = initialOffset;
        if(_align == 2)
            i = (ConversionTemplate.maxTwipsWidth - initialOffset - widthInTwips) / 2;
        else
        if(_align == 1)
            i = ConversionTemplate.maxTwipsWidth - widthInTwips;
        if(i <= 0)
            i = 0;
        int j = 0;
        for(Enumeration enumeration = rows.elements();
			enumeration.hasMoreElements();
			output.write("\\intbl\\row\n"))
        {
            j++;
            Row row = (Row)enumeration.nextElement();
            output.write("\\trowd");
            if (i > 0)
			{
                if (hasBorder)
                	output.write(RtfWriter.TABLE_CELL_BORDER);
                output.write("\n\\cellx" + i + " ");
			}
            int k = i;
            int l = 0;
            for (int j1 = 0; j1 < row.cellCount(); j1++)
            {
                Cell cell = row.getCell(j1);
                if(cell.rowSpanSpacer)
                    output.write("\\clvmrg");
                output.write(cell._bgcolor);
                for(int i1 = 0; i1 < cell.colSpan; i1++)
                    try
                    {
                        if(l + i1 < cellSizes.length)
                            k += cellSizes[l + i1];
                    }
                    catch(IndexOutOfBoundsException _ex) { }

                if (hasBorder)
                	output.write(RtfWriter.TABLE_CELL_BORDER);
                output.write("\n\\cellx" + k);
                l += cell.colSpan;
            }

            output.write("\\pard\n");
            if(i > 0)
                output.write("\\intbl\\cell\\pard\n");
            for (int j2 = 0; j2 < row.cellCount(); j2++)
            {
                Cell cell1 = row.getCell(j2);
                if(hasBorder)
                    output.write("\\intbl\\box ");
                else
                    output.write("\\intbl ");
                if(cell1.isHeader)
                {
                    output.write("{\\fs10 \\line}");
                    output.write("{\\b \\qc ");
                }
                switch(cell1._align)
                {
                case RtfGeneratorContext.ALIGN_RIGHT:
                    output.write("\\qr ");
                    break;

                case RtfGeneratorContext.ALIGN_CENTER:
                    output.write("\\qc ");
                    break;

                case RtfGeneratorContext.ALIGN_LEFT:
                    output.write("\\ql ");
                    break;

                default:
                    if(!cell1.isHeader)
                        output.write("\\ql ");
                    break;
                }
                String s = cell1.contents.toString().trim();
                System.out.println("[RtfTable.dumpRTF]: " + s);
                while (s.endsWith("\\par}")) {
                  int j1 = s.lastIndexOf("\\par}");
                  s = s.substring(0,j1).trim() + "}";
                }
                System.out.println("[RtfTable.dumpRTF]: " + s);
                int j1;
                if(hasBorder)
                    while((j1 = s.indexOf("\\par")) != -1)
                        s = s.substring(0, j1) + "\\line" + s.substring(j1 + 4);
                output.write(s);
                output.write("\\cell\\pard\n");
                if(cell1.isHeader)
                    output.write("}");
            }

        }
    }

  private int textSizeInTwips(String token, int size, boolean bold, boolean italic, boolean fixed)
  {
    return RtfTextSize.textSizeInTwips(token,size,bold,italic,fixed);
  }

    private static final int MARGIN = 300;
    private Vector rows;
    private Row currentRow;
    private Cell currentCell;
    private boolean hasBorder;
    private int widthInTwips;
    private int _align;
    private String _bgcolor;
    private int maxCellPerRow;
    private boolean explicitTableSize;
    private int cellSizes[];
    private int longestNonBreaking[];
    private int maxLineSize[];
    private boolean isEmpty[];
    private boolean explicitSize[];
    private int totalSize;
    private int initialOffset;
    private boolean dumpToCurrentRow;
    private RtfWriter output;
}